"""参数检测。"""
import copy

from simplegeneric import generic

from gm_apidesc.types import SenceType
from gm_apidesc.types import UserType
from gm_apidesc.types import PASS
from gm_apidesc import runexc
from .metatype import _BaseMeta

from ._sence_check import django_input_check

##############################  相关的schema定义检测  ##############################

def register_type(checker, sence_type):
    """场景类型检测"""

    if not sence_type:
        sence_type = SenceType.COMMON

    if sence_type not in SenceType:
        raise runexc.NotMatchError(sence_type, SenceType, "场景类型错误")

    checker.type = sence_type


def register_param(checker, params):
    """参数定义检测"""

    if not isinstance(params, dict):
        raise runexc.SchemaFormatError("schema.param should be dict")

    for param_name, values in params.items():
        if not isinstance(values, (dict, UserType)):
            raise runexc.SchemaFormatError("schema.param format error")

        if isinstance(values, UserType):
            checker.add_type_check(param_name, param_type=values)

        elif isinstance(values, dict):
            param_type = values.get("type")
            if not param_type:
                raise runexc.SchemaFormatError(
                    "schema.param: {param_name} missing 'type'".format(param_name=param_name))

            values.pop("type")
            checker.add_type_check(param_name, param_type=param_type, **values)


def register_return(checker, returns):
    """返回定义检测"""
    pass


def register_desc(checker, desc):
    """描述检测"""
    pass


##############################  相关运行时检测  ##############################

def validate_input(checker, args, kwargs):
    """检测入参。

    :param checker:参数检测对象
    :param args:位置参数值
    :param kwargs:键值对参数值
    """

    args_dict, req_params = combine_args(checker, args, kwargs)

    #  参数名和其他参数名不能冲突
    intersection = set(req_params.keys()) & set(args_dict.keys())
    if intersection:
        raise runexc.ValidateError(
            "req.params names : %s conflict with others" % list(
                intersection))

    for param_name, param_def in checker.param_defs.items():
        # request.POST/GET 参数需要进行转换
        if param_name in req_params:
            validate_with_exception_info(param_def, req_params, param_name)
        else:
            validate_with_exception_info(param_def, args_dict, param_name)

    args_dict.update(req_params)
    # 检查所入参是否检测到
    for arg_name in args_dict:
        if arg_name not in checker.param_defs:
            raise runexc.VisibleExUnknownArgument(arg_name)


def validate_output(checker, return_type, result):
    """出参检测"""
    pass


def combine_args(checker, args, kwargs):
    """将函数参数名称与参数值进行组合。

    :param args: 位置参数值
    :param kwargs: 键值对参数值
    :return:
    """
    args, req_param = checker.get_req_args(args, kwargs)
    kwargs = copy.deepcopy(kwargs)

    args_value = args
    param_names = checker.param_names

    # combine args and kwargs and req.params
    args = dict(zip(param_names, args_value))
    if kwargs:
        args.update(kwargs)

    return args, req_param


def validate_with_exception_info(param_def, dict_param, param_name):
    """
    参数检测入口，提供异常信息的处理
    :param param_def: 参数定义
    :param dict_param: 参数所在的字典结构
    :param param_name:参数名
    :return:
    """
    try:
        check_param(param_def, dict_param, param_name)
    except (runexc.VisibleExInvalidInput,
            runexc.VisibleExMissingArgument,
            runexc.VisibleExUnknownArgument):
        raise
    except Exception as e:
        param_value = dict_param[param_name]
        raise runexc.VisibleExInvalidInput(field=param_name,
                                        value=param_value,
                                        expect_cls=None,
                                        error=e)


def check_param(typedef, value, key=None):
    """检测参数值。

    :param typedef: 参数类型
    :param value: 参数值所在的json
    :param key: 参数对应的key
    :return:
    """

    if key:
        check_typedef(typedef, value, key)
    else:
        datatype = typedef.type
        verification(datatype, value)


def check_required(typedef, value, key):
    if key not in value:
        if typedef.required:
            raise runexc.VisibleExMissingArgument(argname=key, value=value)
        else:
            return PASS


def check_nullable(typedef, value, key):
    """nullable包含None和空字符串。"""

    if typedef.nullable:
        if value[key] in ('', None):
            return PASS

    if not typedef.nullable:

        if value[key] in ('', None):
            raise runexc.VisibleExInvalidInput(field=key,
                                            value=value,
                                            expect_cls=typedef,
                                            error='%s不能为null' % key)

    return


CHECK_ITEM = [check_required, check_nullable]


def check_typedef(typedef, value, key):
    """
    对参数定义项进行检测, 如nullable required
    :param typedef:
    :param value:
    :param key:
    :return:
    """
    for item in CHECK_ITEM:
        if item(typedef, value, key) == PASS:
            return

    from_param(typedef.type, value, key)


# todo 用户自定义类和字典类统一提供validate接口方法
def verification(datatype, value):
    if isinstance(datatype, _BaseMeta):
        check_complex_type(datatype, value)
    else:
        datatype.verification(value)


@generic
def from_param(datatype, value, key):
    raise runexc.VisibleExInvalidInput(field=key,
                                    value=value,
                                    expect_cls=datatype,
                                    error='参数错误')


@from_param.when_type(_BaseMeta)
def from_baseMeta(datatype, value, key):
    check_complex_type(datatype, value[key])


@from_param.when_type(UserType)
def from_UserType(datatype, value, key):
    datatype.type_check(value, key)


def check_complex_type(datatype_def, jdata):
    """检测json类型。"""

    if not isinstance(jdata, dict):
        raise runexc.ValidateError('参数格式错误')

    checker = getattr(datatype_def, "CHECKER_TAG")
    for attr_name in checker.attr_names:
        attr_def = checker.get_elem_def(attr_name)

        try:
            check_param(attr_def, jdata, attr_name)
        except (runexc.VisibleExInvalidInput,
                runexc.VisibleExMissingArgument,
                runexc.VisibleExUnknownArgument):
            raise
        except Exception as e:
            if e.__class__.__name__.startswith('VisibleEx'):
                raise runexc.VisibleExInvalidInput(field=attr_name,
                                                value=jdata,
                                                expect_cls=attr_def.type,
                                                error=e)
            else:
                raise runexc.VisibleExInvalidInput(field=attr_name,
                                                value=jdata,
                                                expect_cls=attr_def.type,
                                                error='参数错误')

    extra_value = []
    for key in jdata:
        if key not in checker.attr_names:
            extra_value.append(key)

    if extra_value:
        raise runexc.VisibleExUnknownArgument(argname=extra_value)

    return jdata



