
from datetime import datetime
import string
import abc

from gm_apidesc import runexc

PASS = True


def filter_inner_type(datatype):
    """将内置类型转换为自定义类型。"""

    if datatype in (int, ):
        return G_IntegerType

    if datatype == str:
        return G_StringType

    if datatype == float:
        return G_FloatType

    if datatype == datetime:
        return Datetime

    if isinstance(datatype, list):
        return ListType(datatype[0])

    if isinstance(datatype, UserType):
        return datatype

    raise runexc.ValidateError("type %s do not allowed" % datatype)


class UserType(object):
    name = None
    whitespace = tuple(string.whitespace)

    def type_check(self, jdata, key):
        self.verification(jdata[key])

    def verification(self, value):
        # 如果字符串前后包含空字符报错
        if type(value) in (str, ):
            if value.endswith(self.whitespace) or value.startswith(
                    self.whitespace):
                raise runexc.ValidateError(
                    "String Type do not allowed start or end with space")
        self.validate(value)

    @abc.abstractmethod
    def validate(self, value):
        pass

    def sample(self):
        # 示例
        pass

    def tostring(self):
        pass

    def set_return_type(self, value):
        pass


class IntegerType(UserType):
    name = "Integer"

    allows_types = (int, str)

    def __init__(self, minimum=None, maximum=None):
        self.minimum = minimum
        self.maximum = maximum

    def validate(self, value):

        if not isinstance(value, self.allows_types) or \
            (isinstance(value, str) and not value.isdigit()):
            raise runexc.VisibleExValueError('值必须是数字')

        value = int(value)

        if self.minimum is not None and value < self.minimum:
            error = '值必须大于等于%s' % self.minimum
            raise runexc.VisibleExValueError(error)

        if self.maximum is not None and value > self.maximum:
            error = '值必须小于等于%s' % self.maximum
            raise runexc.VisibleExValueError(error)


class FloatType(IntegerType):

    name = "Float"

    allows_types = (int, float, str)

    def validate(self, value):

        if not isinstance(value, self.allows_types):
            raise runexc.VisibleExValueError('值必须是数字')

        if isinstance(value, str):
            try:
                value = float(value)
            except:
                raise runexc.VisibleExValueError('值必须是数字')

        if self.minimum is not None and value < self.minimum:
            error = '值必须大于等于%s' % self.minimum
            raise runexc.VisibleExValueError(error)

        if self.maximum is not None and value > self.maximum:
            error = '值必须小于等于%s' % self.maximum
            raise runexc.VisibleExValueError(error)


class _StringType(UserType):

    name = "String"

    def __init__(self, minlen=None, maxlen=None):
        self.minlen = minlen
        self.maxlen = maxlen

    def validate(self, value):

        if not isinstance(value, (str,)):
            raise runexc.VisibleExValueError('值必须是字符串')

        str_len = len(value)

        if self.minlen is not None and str_len < self.minlen:
            error = '字符串长度必须大于等于%s' % self.minlen
            raise runexc.VisibleExValueError(error)

        if self.maxlen is not None and str_len > self.maxlen:
            error = '字符串长度必须小于等于%s' % self.maxlen
            raise runexc.VisibleExValueError(error)


G_StringType = _StringType()
G_IntegerType = IntegerType()
G_FloatType = FloatType()


class _MapperDict(dict):
    def __str__(self):
        # 为dict定制打印格式
        format_dict = dict([(k, v.tostring()) for k, v in self.items()])
        return 'Mapper(%s)' % format_dict


class Mapper(UserType):

    name = 'Mapper'

    def __init__(self, keyname, mapper):

        for key, value in mapper.items():
            mapper[key] = filter_inner_type(value)
        self.mapper = _MapperDict(mapper)
        self.keyname = keyname
        self.valuename = None

    def set_valuename(self, valuename):
        self.valuename = valuename

    def type_check(self, jdata, key):
        self.validate(jdata)

    def sample(self):
        keys = self.mapper.keys()
        return self.mapper[keys[0]]

    def tostring(self):
        return self.mapper

    def validate(self, value):
        # 判断是映射类型
        from gm_apidesc.check import verification
        key = value[self.keyname]
        if key in self.mapper:
            verification(self.mapper[key], value[self.valuename])
        else:
            raise runexc.VisibleExMissingArgument(argname=key, value=value)


class ListType(UserType):

    name = 'List'

    def __init__(self, elemtype, minlen=None, maxlen=None):
        self.elemtype = filter_inner_type(elemtype)
        self.maxlen = maxlen
        self.minlen = minlen

    def validate(self, value):
        if not isinstance(value, list):
            raise runexc.VisibleExValueError('值必须是列表')

        if self.minlen is not None and len(value) < self.minlen:
            error = '元素个数必须大于%s' % self.minlen
            raise runexc.VisibleExValueError(error)

        if self.maxlen is not None and len(value) > self.maxlen:
            error = '元素个数必须小于%s' % self.maxlen
            raise runexc.VisibleExValueError(error)

        for elem in value:
            from sf_apidesc.paramchecker.checkparam import verification
            verification(self.elemtype, elem)

    def sample(self):
        return self.elemtype

    def tostring(self):
        return "[%s]" % self.elemtype.tostring()
